
/*****
      CC_MultiSensor.c -
      Implementation of the ZWave Multilevel Sensor class.  This is for
      Version 11 of the command class.  It uses several application 
      functions for hardware-specific details; this file contains the 
      generic commands.

      c 2019-2020 Primordial Machine Vision Systems
*****/

#include <string.h>
#include <ZAF_tx_mutex.h>
#include <ZAF_Common_interface.h>

#define DEBUGPRINT
#include "DebugPrint.h"

#include <CC_MultiSensor.h>


/**** Macros ****/

/***
    CLIP(val, min, max) - evaluate to min or max if value outside bounds,
                          otherwise to val
***/
#define CLIP(val, min, max)   \
	(((val) < (min)) ? (min) : (((val) > (max)) ? (max) : (val)))


/**** Prototypes ****/

received_frame_status_t send_value(ZW_APPLICATION_TX_BUFFER *,
                                   TRANSMIT_OPTIONS_TYPE_SINGLE_EX *,
                                   uint8_t, int32_t, uint8_t, uint8_t);


/**** Command Implementation ****/

REGISTER_CC(COMMAND_CLASS_SENSOR_MULTILEVEL, SENSOR_MULTILEVEL_VERSION_V11,
            handleCCMultilevelSensor);

/***
    handleCCMultilevelSensor:  Respond to incoming CD_SENSOR_MULTILEVEL
                               commands.
    args:                      rxopt - ZWave options on received frame
                               rxcmd - incoming command
                               rxlen - number of bytes in rxcmd
    returns:   success/failure/unsupported flag
***/
received_frame_status_t
handleCCMultilevelSensor(RECEIVE_OPTIONS_TYPE_EX *rxopt,
                         ZW_APPLICATION_TX_BUFFER *rxcmd, uint8_t rxlen) {
	ZAF_TRANSPORT_TX_BUFFER txbuff;       /* transmit buffer */
	ZW_APPLICATION_TX_BUFFER *txcmd;      /* application data within txbuff */
	TRANSMIT_OPTIONS_TYPE_SINGLE_EX *txopt;   /* outgoing options */
	enum EQueueNotifyingStatus status;    /* transmit result */
	uint8_t masks[16];                    /* bit masks for supported sensors */
	uint8_t nmask;                        /* number masks used */
	uint8_t type;                         /* sensor type to handle */
	uint8_t scale;                        /* scale of sensor to pick */
	uint8_t level;                        /* precision/scale/size of value */
	int32_t val;                          /* value from sensor */
	uint8_t txlen;                        /* length of transmit buffer */

	txcmd = &(txbuff.appTxBuf);
	memset(txcmd, 0, sizeof(*txcmd));
	RxToTxOptions(rxopt, &txopt);

	switch (rxcmd->ZW_Common.cmd) {
	case SENSOR_MULTILEVEL_GET_V11:
		if (Check_not_legal_response_job(rxopt)) {
			DPRINTF("  SENSOR_MULTILEVEL_GET not legal response!\n");
			return RECEIVED_FRAME_STATUS_FAIL;
		}

		/* Per spec, first two commands verify the validity of the type and
		   scale, substituting default values if not. */
		type = rxcmd->ZW_SensorMultilevelGetV11Frame.sensorType;
		scale = (rxcmd->ZW_SensorMultilevelGetV11Frame.properties1 >> 3) & 0x03;
		val = appSensorValue(type, scale);
		level = appSensorLevel(type, scale);

		return send_value(txcmd, txopt, appSensorValidType(type), val, level, 1);
	case SENSOR_MULTILEVEL_SUPPORTED_GET_SENSOR_V11:
		if (Check_not_legal_response_job(rxopt)) {
			DPRINTF("  SENSOR_MULTILEVEL_SUPPORTED_GET_SENSOR not legal response!\n");
			return RECEIVED_FRAME_STATUS_FAIL;
		}

		appSensorMask(masks, &nmask);
		switch (nmask) {
		case 1:
			txcmd->ZW_SensorMultilevelSupportedSensorReport1byteV11Frame.cmdClass =
				COMMAND_CLASS_SENSOR_MULTILEVEL_V11;
			txcmd->ZW_SensorMultilevelSupportedSensorReport1byteV11Frame.cmd =
				SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_V11;
			txcmd->ZW_SensorMultilevelSupportedSensorReport1byteV11Frame.bitMask1 =
				masks[0];
			txlen = sizeof(ZW_SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_1BYTE_V11_FRAME);
			break;
		case 2:
			txcmd->ZW_SensorMultilevelSupportedSensorReport2byteV11Frame.cmdClass =
				COMMAND_CLASS_SENSOR_MULTILEVEL_V11;
			txcmd->ZW_SensorMultilevelSupportedSensorReport2byteV11Frame.cmd =
				SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_V11;
			txcmd->ZW_SensorMultilevelSupportedSensorReport2byteV11Frame.bitMask1 =
				masks[0];
			txcmd->ZW_SensorMultilevelSupportedSensorReport2byteV11Frame.bitMask2 =
				masks[1];
			txlen = sizeof(ZW_SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_2BYTE_V11_FRAME);
			break;
		case 3:
			txcmd->ZW_SensorMultilevelSupportedSensorReport3byteV11Frame.cmdClass =
				COMMAND_CLASS_SENSOR_MULTILEVEL_V11;
			txcmd->ZW_SensorMultilevelSupportedSensorReport3byteV11Frame.cmd =
				SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_V11;
			txcmd->ZW_SensorMultilevelSupportedSensorReport3byteV11Frame.bitMask1 =
				masks[0];
			txcmd->ZW_SensorMultilevelSupportedSensorReport3byteV11Frame.bitMask2 =
				masks[1];
			txcmd->ZW_SensorMultilevelSupportedSensorReport3byteV11Frame.bitMask3 =
				masks[2];
			txlen = sizeof(ZW_SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_3BYTE_V11_FRAME);
			break;
		case 4:
			txcmd->ZW_SensorMultilevelSupportedSensorReport4byteV11Frame.cmdClass =
				COMMAND_CLASS_SENSOR_MULTILEVEL_V11;
			txcmd->ZW_SensorMultilevelSupportedSensorReport4byteV11Frame.cmd =
				SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_V11;
			txcmd->ZW_SensorMultilevelSupportedSensorReport4byteV11Frame.bitMask1 =
				masks[0];
			txcmd->ZW_SensorMultilevelSupportedSensorReport4byteV11Frame.bitMask2 =
				masks[1];
			txcmd->ZW_SensorMultilevelSupportedSensorReport4byteV11Frame.bitMask3 =
				masks[2];
			txcmd->ZW_SensorMultilevelSupportedSensorReport4byteV11Frame.bitMask4 =
				masks[3];
			txlen = sizeof(ZW_SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_4BYTE_V11_FRAME);
			break;
		default:
			DPRINTF("sensor mask needs %d bytes but ZWave API only allows up to 4\n",
			        nmask);
			return RECEIVED_FRAME_STATUS_FAIL;
		}
		break;
	case SENSOR_MULTILEVEL_SUPPORTED_GET_SCALE_V11:
		if (Check_not_legal_response_job(rxopt)) {
			DPRINTF("  SENSOR_MULTILEVEL_SUPPORTED_GET_SCALE not legal response!\n");
			return RECEIVED_FRAME_STATUS_FAIL;
		}

		type = rxcmd->ZW_SensorMultilevelSupportedGetScaleV11Frame.sensorType;

		txcmd->ZW_SensorMultilevelSupportedScaleReportV11Frame.cmdClass =
			COMMAND_CLASS_SENSOR_MULTILEVEL_V11;
		txcmd->ZW_SensorMultilevelSupportedScaleReportV11Frame.cmd =
			SENSOR_MULTILEVEL_SUPPORTED_SCALE_REPORT_V11;
		txcmd->ZW_SensorMultilevelSupportedScaleReportV11Frame.sensorType = type;
		txcmd->ZW_SensorMultilevelSupportedScaleReportV11Frame.properties1 =
			appSensorScaleMask(type);
		txlen = sizeof(ZW_SENSOR_MULTILEVEL_SUPPORTED_SCALE_REPORT_V11_FRAME);
		break;
	case SENSOR_MULTILEVEL_REPORT_V11:
		/* Should never get an incoming report. */
		DPRINTF("  SENSOR_MULTILEVEL_REPORT not supported\n");
		return RECEIVED_FRAME_STATUS_NO_SUPPORT;
	case SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT_V11:
		DPRINTF("  SENSOR_MULTILEVEL_SUPPORTED_SENSOR_REPORT not supported\n");
		return RECEIVED_FRAME_STATUS_NO_SUPPORT;
	case SENSOR_MULTILEVEL_SUPPORTED_SCALE_REPORT_V11:
		DPRINTF("  SENSOR_MULTILEVEL_SUPPORTED_SCALE_REPORT not supported\n");
		return RECEIVED_FRAME_STATUS_NO_SUPPORT;
	default:
		DPRINTF("  unknown SENSOR_MULTILEVEL command 0x%02x\n", rxcmd->ZW_Common.cmd);
		return RECEIVED_FRAME_STATUS_NO_SUPPORT;
	}

	status = Transport_SendResponseEP((uint8_t *) txcmd, txlen, txopt, NULL);
	if (EQUEUENOTIFYING_STATUS_SUCCESS == status) {
		return RECEIVED_FRAME_STATUS_SUCCESS;
	} else {
		return RECEIVED_FRAME_STATUS_FAIL;
	}
}

/***
    sendSensorValue:  Pack the sensor reading into a report and send it off.
    args:             type - sensor type
                      val - sensor reading
                      level - value's precision, scale, and size packed
    returns:   success/failure/unsupported result
***/
received_frame_status_t sendSensorValue(uint8_t type, int32_t val,
                                        uint8_t level) {
	ZAF_TRANSPORT_TX_BUFFER txbuff;       /* transmit buffer */
	ZW_APPLICATION_TX_BUFFER *txcmd;      /* application data within txbuff */
	TRANSMIT_OPTIONS_TYPE_SINGLE_EX *txopt;   /* outgoing options */
	RECEIVE_OPTIONS_TYPE_EX rxopt;            /* dummy incoming options */

	txcmd = &(txbuff.appTxBuf);
	memset(txcmd, 0, sizeof(*txcmd));

	memset(&rxopt, 0, sizeof(rxopt));
	/* Fake an incoming packet from the host, assuming its encrypted if
	   we have an S0 security key.  This would eventually go to the
	   lifeline. */
	rxopt.sourceNode.nodeId = 1;
	if (ZAF_GetSecurityKeys() & SECURITY_KEY_S0_BIT) {
		rxopt.securityKey = SECURITY_KEY_S0;
	}
	RxToTxOptions(&rxopt, &txopt);
	
	return send_value(txcmd, txopt, type, val, level, 0);
}



/**** Support ****/

/***
    send_value:  Build the appropriate report depending on the size of the
                 value packed into the level and send it off.
    args:        txcmd - transmit buffer
                 txopt - transmit options
                 type - sensor type
                 val - sensor value (will clip to size if needed)
                 level - value's precision, scale, and size packed
                 isRES - true to send through response queue, 0 request
    returns:   sucess/failure/unsupported result
***/
received_frame_status_t send_value(ZW_APPLICATION_TX_BUFFER *txcmd,
                                   TRANSMIT_OPTIONS_TYPE_SINGLE_EX *txopt,
                                   uint8_t type, int32_t val, uint8_t level,
                                   uint8_t isRES) {
	enum EQueueNotifyingStatus status;    /* transmit result */
	int16_t sval;                         /* 2 byte value */
	int8_t cval;                          /* 1 byte value */
	uint8_t txlen;                        /* length of transmit buffer */

	switch (level & 0x07) {
	case 1:
		cval = CLIP(val, -128, 127);
		txcmd->ZW_SensorMultilevelReport1byteV11Frame.cmdClass =
			COMMAND_CLASS_SENSOR_MULTILEVEL_V11;
		txcmd->ZW_SensorMultilevelReport1byteV11Frame.cmd =
			SENSOR_MULTILEVEL_REPORT_V11;
		txcmd->ZW_SensorMultilevelReport1byteV11Frame.sensorType = type;
		txcmd->ZW_SensorMultilevelReport1byteV11Frame.level = level;
		txcmd->ZW_SensorMultilevelReport1byteV11Frame.sensorValue1 = cval & 0xff;
		txlen = sizeof(ZW_SENSOR_MULTILEVEL_REPORT_1BYTE_V11_FRAME);
		break;
	case 2:
		sval = CLIP(val, -32768, 32767);
		txcmd->ZW_SensorMultilevelReport2byteV11Frame.cmdClass =
			COMMAND_CLASS_SENSOR_MULTILEVEL_V11;
		txcmd->ZW_SensorMultilevelReport2byteV11Frame.cmd =
			SENSOR_MULTILEVEL_REPORT_V11;
		txcmd->ZW_SensorMultilevelReport2byteV11Frame.sensorType = type;
		txcmd->ZW_SensorMultilevelReport2byteV11Frame.level = level;
		txcmd->ZW_SensorMultilevelReport2byteV11Frame.sensorValue1 =
			(sval >> 8) & 0xff;
		txcmd->ZW_SensorMultilevelReport2byteV11Frame.sensorValue2 = sval & 0xff;
		txlen = sizeof(ZW_SENSOR_MULTILEVEL_REPORT_2BYTE_V11_FRAME);
		break;
	case 4:
		txcmd->ZW_SensorMultilevelReport4byteV11Frame.cmdClass =
			COMMAND_CLASS_SENSOR_MULTILEVEL_V11;
		txcmd->ZW_SensorMultilevelReport4byteV11Frame.cmd =
			SENSOR_MULTILEVEL_REPORT_V11;
		txcmd->ZW_SensorMultilevelReport4byteV11Frame.sensorType = type;
		txcmd->ZW_SensorMultilevelReport4byteV11Frame.level = level;
		txcmd->ZW_SensorMultilevelReport4byteV11Frame.sensorValue1 =
			(val >> 24) & 0xff;
		txcmd->ZW_SensorMultilevelReport4byteV11Frame.sensorValue2 =
			(val >> 16) & 0xff;
		txcmd->ZW_SensorMultilevelReport4byteV11Frame.sensorValue3 =
			(val >> 8) & 0xff;
		txcmd->ZW_SensorMultilevelReport4byteV11Frame.sensorValue4 = val & 0xff;
		txlen = sizeof(ZW_SENSOR_MULTILEVEL_REPORT_4BYTE_V11_FRAME);
		break;
	default:
		return RECEIVED_FRAME_STATUS_FAIL;
	}

	if (isRES) {
		status = Transport_SendResponseEP((uint8_t *) txcmd, txlen, txopt, NULL);
	} else {
		status = Transport_SendRequestEP((uint8_t *) txcmd, txlen, txopt, NULL);
	}
	if (EQUEUENOTIFYING_STATUS_SUCCESS == status) {
		return RECEIVED_FRAME_STATUS_SUCCESS;
	} else {
		return RECEIVED_FRAME_STATUS_FAIL;
	}
}

