
/*****
      HTSensor3E.c -
      Add Configuration CC for setting and controlling data logging.

			c 2019-2020 Primordial Machine Vision Systems
*****/

#include <string.h>

#include "board.h"
#include "i2cspm.h"
#include "display.h"
#include "textdisplay.h"
#include "graphdisplay.h"

#include "AppTimer.h"
#define DEBUGPRINT
#include "DebugPrint.h"
#include "DebugPrintConfig.h"
#include "EventDistributor.h"
#include "SizeOf.h"

#include "ZW_application_transport_interface.h"
#include "ZW_basis_api.h"
#include "ZW_classcmd.h"
#include "ZW_system_startup_api.h"
#include "ZW_TransportSecProtocol.h"

#include "ZAF_Common_helper.h"
#include "zaf_event_helper.h"
#include "ZAF_network_learn.h"
#include "ZAF_TSE.h"
#include "ZAF_types.h"
#include "ZAF_uart_utils.h"
 
#include "CC_Version.h"
#include "CC_MultiSensor.h"
#include "CC_Configuration.h"

#include "config_app.h"
#include "config_rf.h"

#include "FreeRTOS.h"
#include "semphr.h"


/**** Constants - Application ****/

/* maximum number of events we can store */
#define MAXEVENT                   5

/* max number of samples in a data logging run
   (in practice, limit to width of drawing area) */
#define NLOG                      64

/* time in seconds between data logs */
#define TSAMPLE                    5

/* height of graph in pixels */
#define GRAPH_HT                  64

/* top line/row of graph on display */
#define GRAPH_TOP                 46

/* minimum range (max-min) for graph vertical interval, in milli-%/C */
/**!!!**/
/* small for test */
#define GRAPH_MINRANGE           500

/* time in ms to wait for lock on lock */
#define TLOCK                    100

/* unsecured command classes */
static uint8_t cc_unsecure[] = {
	COMMAND_CLASS_VERSION,
	COMMAND_CLASS_SECURITY
};

/* secured command classes available unsecured */
static uint8_t cc_open_secure[] = {
	COMMAND_CLASS_SECURITY
};

/* secured command classes */
static uint8_t cc_secure[] = {
	COMMAND_CLASS_SENSOR_MULTILEVEL,
	COMMAND_CLASS_CONFIGURATION,
	COMMAND_CLASS_SECURITY
};

/* first node info setup (must be same info as appnif2 - use copy_nif() */
static app_node_information_t appnif1 = {
	cc_unsecure, sizeof_array(cc_unsecure),
	cc_open_secure, sizeof_array(cc_open_secure),
	cc_secure, sizeof_array(cc_secure),
	DEVICE_OPTIONS_MASK, {GENERIC_TYPE, SPECIFIC_TYPE}
};

/**** Constants - Internal ****/

/* Our supported sensor types.  The icon code contains the value in the low
   byte, otherwise we'd have to manually copy from SDS13812. */
#define SENSOR_TEMP   (ICON_TYPE_SPECIFIC_SENSOR_MULTILEVEL_AIR_TEMPERATURE & 0xff)
#define SENSOR_RH     (ICON_TYPE_SPECIFIC_SENSOR_MULTILEVEL_HUMIDITY & 0xff)

/* Our supported sensor scales.  Defined in SDS13812. */
#define SCALE_DEGC            0x00
#define SCALE_DEGF            0x01
#define SCALE_PCT             0x00

/* logging configuration parameters */
#define CONFIG_TSAMPLE           1
#define CONFIG_NSAMPLE           2
#define CONFIG_RUN               3

/* sensor commands - see Si7021 spec chapter 5 */
#define READ_H_HOLD           0xE5
#define READ_H_NOHOLD         0xF5
#define READ_T_HOLD           0xE3
#define READ_T_NOHOLD         0xF3
#define READ_T_AFTER_H        0xE0
#define TH_RESET              0xFE
#define WRITE_USER            0xE6
#define READ_USER             0xE7
#define WRITE_HEATER          0x51
#define READ_HEATER           0x11
#define READ_ID1_BYTE1        0xFA
#define READ_ID1_BYTE2        0x0F
#define READ_ID2_BYTE1        0xFC
#define READ_ID2_BYTE2        0xC9
#define READ_FW_BYTE1         0xB4
#define READ_FW_BYTE2         0xB8

/* events that 'can wake up protocol thread' */
enum evtIDs {
	EVT_TIMER = 0,
	EVT_ZWRX,
	EVT_ZWCMD,
	EVT_APP
};

enum zwevtIDs {
	EVT_EMPTY = 0
};

/* network include status */
enum netstate {
	LEARN_EXCLUDED,
	LEARN_INCLUDED,
	LEARN_UNDERWAY,
	LEARN_FAILED
};

/* EMSTATUS normally encodes a base value - just a simple fail (> 0) value */
#define EMSTATUS_FAIL            1


/**** Structures ****/

/* sensor address on I2C bus - see datasheet, section 5 */
#define TH_I2CADDR            0x80

/* repeated samples and graph setup */
struct htsamples {
	bool running;                         /* true when logger active */
	uint8_t nsample;                      /* number samples to take */
	uint8_t tsample;                      /* time in seconds between data logs */
	uint8_t id;                           /* current sample index */
	uint32_t rh[NLOG];                    /* humidity (milli-percent) */
	uint32_t rhmin;                       /* lowest humidity in log */
	uint32_t rhmax;                       /* largest humidity in log */
	uint32_t g0min;                       /* lower humidity bound graph 0 */
	uint32_t g0scl;                       /* scaling humidity graph 0 */
	int32_t tC[NLOG];                     /* temperature (milli-degrees C) */
	int32_t tCmin;                        /* lowest temperature in log */
	int32_t tCmax;                        /* highest temperature in log */
	int32_t g1min;                        /* lower temperature bound graph 1 */
	int32_t g1scl;
};

struct appconfig {
	DISPLAY_Device_t lcd;                 /* overall display setup */
	TEXTDISPLAY_Handle_t text;            /* text half of LCD */
	GRAPHDISPLAY_Handle_t graph;          /* graphical half of LCD */
	bool learning;                        /* true when in/excluding */
};


/**** Globals ****/

/* humidity, temperature logs */
static struct htsamples htlog;

/* mutex protecting access to log, and its reference handle */
static StaticSemaphore_t htlog_mutex;
static SemaphoreHandle_t htlog_lock;

/* setup/state of this application */
static struct appconfig app;

/* timer for sampling */
static SSwTimer sampler;

/* FreeRTOS queue holding application events */
static QueueHandle_t evtqueue;

/* overal ZW configuration */
static SProtocolConfig_t zwcfg;

/* ZW node info setup in another structure (copy from appnif1) */
static SAppNodeInfo_t appnif2;

/* radio setup */
static SRadioConfig_t radio;

/* security keys setup */
static uint8_t s2keys;

#ifdef DEBUGPRINT
/* buffer to hold print output - note one line! */
static uint8_t dbgbuff[96];
#endif   /* DEBUGPRINT */


/**** Macros ****/

/* Return a non-zero error code (ie. not EMSTATUS_OK). */
#define RET_ON_ERR(err)                     \
	do { if (EMSTATUS_OK != (err)) { return (err); }; } while (0)

/* Try to grab the mutex on htlog, waiting up to TLOCK ms.  If it fails jump
   to the label 'onerror'.  lock is the mutex handle. */
#define END_ON_LOCKERR(lock)                \
	do { if (pdPASS != xSemaphoreTake(lock, pdMS_TO_TICKS(TLOCK))) { DPRINTF("lock error in %s (%d)\n", __FUNCTION__, __LINE__); goto onerror; } } while (0)

/* Try to release the mutex on htlog.  If it fails jump to the label 'onerror'.
   lock is the mutex handle. */
#define END_ON_UNLOCKERR(lock)              \
	do { if (pdPASS != xSemaphoreGive(lock)) { DPRINTF("unlock error in %s (%d)\n", __FUNCTION__, __LINE__); goto onerror; } } while (0)

/* Release the mutex lock on htlog, ignoring errors */
#define UNLOCK(lock)                        \
	do { xSemaphoreGive(lock); } while (0)


/**** Prototypes ****/

static void handle_event(void);
static void start_sampling(void);
static void sample_ht(SSwTimer *);

static void clear_log(void);
static void stop_log(SSwTimer *);
static EMSTATUS bound_nsample(uint8_t *);
static void setup_i2c(void);
static const char *i2c_strerror(I2C_TransferReturn_TypeDef);
static void verify_sensor(void);
static uint32_t read_HT(uint32_t *, int32_t *);

static uint32_t setup_display(void);
static EMSTATUS display_humidity(uint32_t);
static EMSTATUS display_temperature(int32_t);
static void clear_graphs(void);
static EMSTATUS graph_humidity(void);
static EMSTATUS graph_temperature(void);
static EMSTATUS draw_2x2point(uint8_t, uint8_t, uint8_t, bool);

static received_frame_status_t send_temperature(int32_t);
static received_frame_status_t send_humidity(uint32_t);

static void set_indicator(enum netstate);

static void ApplicationTask(SApplicationHandles *);

static void handle_frame(void);
static void handle_cmdstatus(void);

static bool am_included(void);
static void copy_nif(app_node_information_t *, SAppNodeInfo_t *);



/**** Event Loop - Application ****/

/***
    handle_event:  Process all events in the queue.  The main hardware
                   handler.
    modifies:  htlog
***/
static void handle_event(void) {
	SApplicationHandles *zwapp;           /* ZWave configuration and queues */
	enum zwevtIDs evt;                    /* an event in the queue */
	uint32_t rh;                          /* relative humidity measurement */
	int32_t tC;                           /* temperature measurement */
	uint32_t err;

	while (xQueueReceive(evtqueue, (void *) &evt, 0)) {
		/* Despite BTN_EVENT_SHORT_PRESS being a macro (see board.h), it calls a
		   function so we can't use a switch statement here. */
		if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB1) == (BUTTON_EVENT) evt) {
			if (app.learning) {
				/**!!!**/
				/* You could cancel the learn ... */
				continue;
			}
			zwapp = ZAF_getAppHandle();
			if (EINCLUSIONSTATE_EXCLUDED == zwapp->pNetworkInfo->eInclusionState) {
				ZAF_setNetworkLearnMode(E_NETWORK_LEARN_MODE_INCLUSION,
				                        ERESETREASON_OTHER);
			} else {
				ZAF_setNetworkLearnMode(E_NETWORK_LEARN_MODE_EXCLUSION_NWE,
				                        ERESETREASON_OTHER);
			}
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB2) == (BUTTON_EVENT) evt) {
			END_ON_LOCKERR(htlog_lock);
			if (0 == htlog.nsample) {
				Board_SetLed(BOARD_LED2, LED_ON);
				htlog.running = 1;
				htlog.nsample = 0;
				err = bound_nsample(&(htlog.nsample));
				if (EMSTATUS_OK != err) {
					UNLOCK(htlog_lock);
					return;
				}
				htlog.tsample = TSAMPLE;
				END_ON_UNLOCKERR(htlog_lock);
				DPRINTF(" starting log of %u samples every %u s\n", htlog.nsample, htlog.tsample);
				start_sampling();
			} else {
				htlog.nsample = 0;
				END_ON_UNLOCKERR(htlog_lock);
			}
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB3) == (BUTTON_EVENT) evt) {
			Board_SetLed(BOARD_LED4, LED_ON);
			err = read_HT(NULL, &tC);
			if (0 == err) {
				DPRINTF("  temperature  %3d.%03d C\n", tC/1000, ((tC<0) ? -tC : tC)%1000);
				display_temperature(tC);
				send_temperature(tC);
			}
			Board_SetLed(BOARD_LED4, LED_OFF);
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB4) == (BUTTON_EVENT) evt) {
			Board_SetLed(BOARD_LED4, LED_ON);
			err = read_HT(&rh, NULL);
			if (0 == err) {
				DPRINTF("  humidity     %3d.%03d %%\n", rh/1000, rh%1000);
				display_humidity(rh);
				send_humidity(rh);
			}
			Board_SetLed(BOARD_LED4, LED_OFF);
		}
	}

onerror:
	return;
}

/***
    start_sampling:  Set up the log and start the software timer.  The 
                     sampling parameters (number of samples, time between)
                     must be already set in htlog.
    modifies:  htlog
***/
static void start_sampling(void) {

	clear_log();
	clear_graphs();

	TimerStart(&sampler, htlog.tsample*1000);
	END_ON_LOCKERR(htlog_lock);
	htlog.running = 1;
	END_ON_UNLOCKERR(htlog_lock);

onerror:
	return;
}

/***
    sample_ht:  Read the humidity and temperature and store the values in
                the log.  Start another timer interval if we're not done.
    args:       timer - clock that waits until it's time for the next sample
    modifies:  htlog
***/
static void sample_ht(SSwTimer *timer) {
	uint32_t rh;                          /* humidity measurement */
	int32_t tC;                           /* temperature measurement */
	uint32_t err;

	END_ON_LOCKERR(htlog_lock);
	if (htlog.id == htlog.nsample) {
		DPRINT("timer still running even though log is full\n");
		UNLOCK(htlog_lock);
		stop_log(timer);
		return;
	}
	/* We signal an early stop to sampling by clearing nsample. */
	if ((0 < htlog.id) && (0 == htlog.nsample)) {
		DPRINT(" logging stopped\n");
		UNLOCK(htlog_lock);
		stop_log(timer);
		return;
	}
	/* Stopped before we started. */
	if (0 == htlog.nsample) {
		UNLOCK(htlog_lock);
		return;
	}
	END_ON_UNLOCKERR(htlog_lock);

	err = read_HT(&rh, &tC);
	if (0 != err) {
		DPRINT(" logging stopped after error reading temperature\n");
		stop_log(timer);
		return;
	}

	END_ON_LOCKERR(htlog_lock);
	htlog.rh[htlog.id] = rh;
	if (0 < rh) {
		if (rh < htlog.rhmin) {
			htlog.rhmin = rh;
		}
		if (htlog.rhmax < rh) {
			htlog.rhmax = rh;
		}
	}

	htlog.tC[htlog.id] = tC;
	if (0 < tC) {
		if (tC < htlog.tCmin) {
			htlog.tCmin = tC;
		}
		if (htlog.tCmax < tC) {
			htlog.tCmax = tC;
		}
	}
	END_ON_UNLOCKERR(htlog_lock);

	err = display_humidity(rh);
	if (EMSTATUS_OK != err) {
		DPRINT(" logging stopped after humidity display error\n");
		stop_log(timer);
		return;
	}
	/* Note: don't let a ZWave error stop our logging. */
	send_humidity(rh);

	err = display_temperature(tC);
	if (EMSTATUS_OK != err) {
		DPRINT(" logging stopped after temperature display error\n");
		stop_log(timer);
		return;
	}
	send_temperature(tC);

	/**!!!**/
	/* Re-work so that graph points are drawn with sending to screen, then
	   copy pixel buffer once when done, rather than twice at the end of
	   each graph_*. */
	err = graph_humidity();
	if (EMSTATUS_OK != err) {
		DPRINT(" logging stopped after humidity graphing error\n");
		stop_log(timer);
		return;
	}
	err = graph_temperature();
	if (EMSTATUS_OK != err) {
		DPRINT(" logging stopped after temperature graphing error\n");
		stop_log(timer);
		return;
	}

	END_ON_LOCKERR(htlog_lock);
	htlog.id += 1;
	if (htlog.nsample <= htlog.id) {
		DPRINT(" logging done\n");
		UNLOCK(htlog_lock);
		stop_log(timer);
		return;
	}
	END_ON_UNLOCKERR(htlog_lock);

onerror:
	return;
}


/**** Sensor/Log Interface ****/

/***
    clear_log:  Initialize the log.  Do not change the nsample and tsample
                members.
    modifies:  htlog
***/
static void clear_log(void) {
	uint8_t nsample;                      /* temp storage of existing nsample */
	uint8_t tsample;                      /* temp storage of existing tsample */

	END_ON_LOCKERR(htlog_lock);

	/* This seems easier than zero all other members one by one. */
	nsample = htlog.nsample;
	tsample = htlog.tsample;
	memset(&htlog, 0, sizeof(htlog));
	htlog.nsample = nsample;
	htlog.tsample = tsample;

	/* Max RH is 100%.  Max storage temp is -65 to +150 C in the spec. */
	htlog.rhmin = 100001;
	htlog.tCmin = 150001;
	htlog.tCmax = -65001;

	END_ON_UNLOCKERR(htlog_lock);

onerror:
	return;
}

/***
    stop_log:  Halt the logging timer and turn off the status LED.
    args:      timer - clock triggering log samples
    modifies:  htlog
***/
static void stop_log(SSwTimer *timer) {

	Board_SetLed(BOARD_LED2, LED_OFF);
	TimerStop(timer);
	htlog.running = 0;
}

/***
    bound_nsample:  Modify the sample value to be a valid number that will fit
                    in the graph.  If the initial number is 0 then set it to
                    the maximum possible.
    args:           nsample - number of samples to take
    returns:   0 if successful, > 0 on error (value depends on error)
    modifies:  nsample
***/
static EMSTATUS bound_nsample(uint8_t *nsample) {
	uint8_t wgraph;                       /* width drawing area = max samples */
	uint32_t err;

	err = GRAPHDISPLAY_GetWidth(app.graph, &wgraph);
	if (EMSTATUS_OK != err) {
		return err;
	}
	if (NLOG < wgraph) {
		wgraph = NLOG;
	}

	if ((0 == *nsample) || (wgraph < *nsample)) {
		*nsample = wgraph;
	}

	return EMSTATUS_OK;
}

/***
    setup_i2c:  Initialize the I2C port for the humidity/temperature sensor.
                Verify that we have a sensor (which will fail an ASSERT if
                we can't find one).
***/
static void setup_i2c(void) {
	I2CSPM_Init_TypeDef i2c0;             /* I2C setup */

	/* Assumes the sensor is already enabled, by setting up the display
	   before this. */

	/* The values come from UG381.  Can't find a match in the hardware
	   kits for our board. */
	i2c0.port = I2C0;
	i2c0.sclPort = gpioPortC;
	i2c0.sclPin = 10;
	i2c0.portLocationScl = 14;
	i2c0.sdaPort = gpioPortC;
	i2c0.sdaPin = 11;
	i2c0.portLocationSda = 16;
	i2c0.i2cRefFreq = 0;
	i2c0.i2cMaxFreq = I2C_FREQ_STANDARD_MAX;
	i2c0.i2cClhr = i2cClockHLRStandard;
	I2CSPM_Init(&i2c0);

	verify_sensor();
}

/***
    i2c_strerror:  Convert the I2C error code to a string.
    args:          errno - result from I2C operation
    returns:   string name for error code
***/
static const char *i2c_strerror(I2C_TransferReturn_TypeDef errno) {

	switch (errno) {
	case i2cTransferInProgress:   return "in progress";
	case i2cTransferDone:         return "done";
	case i2cTransferNack:         return "NACK";
	case i2cTransferBusErr:       return "bus error";
	case i2cTransferArbLost:      return "lost arbitration";
	case i2cTransferUsageFault:   return "usage fault";
	case i2cTransferSwFault:      return "SW fault";
	default:                      return "unknown";
	}
}

/***
    verify_sensor:  Test that we can find the temp/humidity sensor.  Assumes
                    we're looking for an Si7021.  Fails an ASSERT if sensor
                    not found.
***/
static void verify_sensor(void) {
	I2C_TransferSeq_TypeDef tx;           /* packet to transfer */
	I2C_TransferReturn_TypeDef status;    /* message back from sensor */
	uint8_t txbuff[2];                    /* buffer holding outgoing message */
	uint8_t rxbuff[8];                    /* buffer holding incoming message */
	uint8_t ntry;                         /* try counter */
	int8_t found;                         /* true if we found the sensor */

	txbuff[0] = READ_ID2_BYTE1;
	txbuff[1] = READ_ID2_BYTE2;

	/* From datasheet, section 5. */
	tx.addr = TH_I2CADDR;
	tx.flags = I2C_FLAG_WRITE_READ;
	tx.buf[0].data = txbuff;
	tx.buf[0].len = sizeof_array(txbuff);
	tx.buf[1].data = rxbuff;
	tx.buf[1].len = sizeof_array(rxbuff);

	found = 0;
	/* The try cycle was used debugging the I2C setup.  Is there any reason
	   to keep it?  Technically the transfer could stall ... */
	ntry = 2;
	status = i2cTransferInProgress;
	while ((0 < ntry) && (i2cTransferInProgress == status)) {
		status = I2CSPM_Transfer(I2C0, &tx);
		if (i2cTransferDone == status) {
			/* Per spec, 0x0D is a Si7013, 0x14 a Si7020.  These aren't on our
			   devkit board. */
			if (0x15 != rxbuff[0]) {
				DPRINTF("FAILED to find sensor: ID 0x%02x not 7021 ID 0x%02x\n",
				        rxbuff[0], 0x15);
			} else {
				DPRINT("found sensor!\n");
				found = 1;
			}
			break;
		}
		ntry -= 1;
	}
	if ((0 == ntry) && (i2cTransferInProgress == status)) {
		DPRINT("FAILED to find sensor: out of tries\n");
	} else if (i2cTransferDone != status) {
		/* See em_i2c.h for error codes. */
		DPRINTF("FAILED to find sensor: error %d (%s)\n",
		        status, i2c_strerror(status));
	}

	ASSERT(found);
}

/***
    read_HT:  Grab a humidity and temperature measurement from the sensor.
              The temperature is read at the same time as the humidity.
              Pass NULL for an argument to skip that measurement.
    args:     rh - relative humidity (skip if NULL)
              tC - temperature (skip if NULL)
    returns:   0 if successful, > 0 on error
    modifies:  rh, tC
***/
static uint32_t read_HT(uint32_t *rh, int32_t *tC) {
	I2C_TransferSeq_TypeDef tx;           /* packet to transfer */
	I2C_TransferReturn_TypeDef status;    /* message back from sensor */
	uint32_t meas;                        /* measured value */
	uint8_t txbuff[1];                    /* outgoing message */
	uint8_t rxbuff[2];                    /* incoming buffer */
	uint8_t ntry;                         /* try counter */

	if ((NULL == rh) && (NULL == tC)) {
		return 0;
	}

	tx.addr = TH_I2CADDR;
	tx.flags = I2C_FLAG_WRITE_READ;
	tx.buf[0].data = txbuff;
	tx.buf[0].len = sizeof_array(txbuff);
	tx.buf[1].data = rxbuff;
	tx.buf[1].len = sizeof_array(rxbuff);

	if (NULL != rh) {
		txbuff[0] = READ_H_HOLD;
		ntry = 2;
		status = i2cTransferInProgress;
		while ((0 < ntry) && (i2cTransferInProgress == status)) {
			status = I2CSPM_Transfer(I2C0, &tx);
			ntry -= 1;
		}

		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read humidity - out of tries\n");
			*rh = 0;
			return 1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read humidity - error %d (%s)\n",
			        status, i2c_strerror(status));
			*rh = 0;
			return 1;
		}

		meas = ((uint32_t) rxbuff[0] << 8) + (rxbuff[1] & 0xfc);
		*rh = (meas * 125000L) >> 16;
		if (*rh < 6000) {
			*rh = 0;
		} else if (106000 < *rh) {
			*rh = 100000;
		} else {
			*rh -= 6000;
		}
	}

	if (NULL != tC) {
		if (NULL != rh) {
			txbuff[0] = READ_T_AFTER_H;
		} else {
			txbuff[0] = READ_T_HOLD;
		}

		ntry = 2;
		status = i2cTransferInProgress;
		while ((0 < ntry) && (i2cTransferInProgress == status)) {
			status = I2CSPM_Transfer(I2C0, &tx);
			ntry -= 1;
		}

		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read temperature - out of tries\n");
			*tC = 0;
			return 1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read temperature - error %d (%s)\n",
			        status, i2c_strerror(status));
			*tC = 0;
			return 1;
		}
	
		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read temperature - out of tries\n");
			return 1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read temperature - error %d (%s)\n",
			        status, i2c_strerror(status));
			return 1;
		}

		meas = ((uint32_t) rxbuff[0] << 8) + (rxbuff[1] & 0xfc);
		/* The formula in the spec overflows at 32 bit, ie. meas * 175720.
		   Scaling the fraction by 8 above and below. */
		*tC = (meas * 21965) >> 13;
		*tC -= 46850;
	}

	return 0;
}


/**** LCD ****/

/***
    setup_display:  Prepare the display, including text support.  Will stop
                    halfway through process if any step fails.  As a side
                    effect of initializing the LCD, it's also enabled - which
                    jointly enables the sensor.
    returns:   0 if successful, > 0 on error (value depends on error)
    modifies:  app
***/
static uint32_t setup_display(void) {
	TEXTDISPLAY_Config_t textcfg;         /* textdisplay configuration */
	GRAPHDISPLAY_Config_t graphcfg;       /* graphdisplay configuration */
	uint32_t err;

	err = DISPLAY_Init();
	RET_ON_ERR(err);

	err = DISPLAY_DeviceGet(0, &app.lcd);
	RET_ON_ERR(err);

	/* Function only uses these members of cfg: displayDeviceNo,
	   scrollEnable, lfToCrLf. */
	textcfg.displayDeviceNo = 0;
	textcfg.scrollEnable = 0;
	textcfg.lfToCrLf = 0;
	textcfg.maxlines = 2;
	err = TEXTDISPLAY_New(&textcfg, &app.text);
	RET_ON_ERR(err);

	graphcfg.displayDeviceNo = 0;
	graphcfg.y0 = GRAPH_TOP;
	graphcfg.hdraw = GRAPH_HT;
	graphcfg.ngraph = 2;
	err = GRAPHDISPLAY_New(&graphcfg, &app.graph);
	RET_ON_ERR(err);

	return 0;
}

/***
    display_humidity:  Print the humidity (XXX.X %) to the first line of
                       the text display.  The rest of the display isn't
                       changed.
    args:              rh - humidity (milli-percent)
    returns:   0 if successful, > 0 on error (value depends on error)
***/
static EMSTATUS display_humidity(uint32_t rh) {
	uint32_t pwr10;                       /* which digit to print */
	EMSTATUS err;

	/* Maximum width is 7 characters, but since we should never see 100%,
	   the first digit will be a blank.  Therefore left-justify the
	   string so it looks centered, usually. */

	err = TEXTDISPLAY_WriteString(app.text, TEXTDISPLAY_ESC_SEQ_CURSOR_HOME_VT52);
	RET_ON_ERR(err);

	/* Round up the tenths digit. */
	rh += 50;

	pwr10 = 100000;
	while (10 < pwr10) {
		/* Don't print leading zeros before the unit's digit (pwr10==1000),
		   put do print it and after the decimal point. */
		if ((0 == (rh / pwr10)) && (1000 < pwr10)) {
			err = TEXTDISPLAY_WriteChar(app.text, ' ');
			RET_ON_ERR(err);
		} else {
			err = TEXTDISPLAY_WriteChar(app.text, '0' + (rh / pwr10));
			RET_ON_ERR(err);
		}
		if (1000 == pwr10) {
			err = TEXTDISPLAY_WriteChar(app.text, '.');
			RET_ON_ERR(err);
		}

		rh %= pwr10;
		pwr10 /= 10;
	}

	err = TEXTDISPLAY_WriteChar(app.text, ' ');
	RET_ON_ERR(err);

	err = TEXTDISPLAY_WriteChar(app.text, '/');
	RET_ON_ERR(err);

	return EMSTATUS_OK;
}

/***
    display_temperature:  Print the temperature ([-]XXX.X degC) to the 
                          second line of the text display.  The rest of
                          the display isn't changed.
    args:                 tC - temperature (milli-degree C)
    returns:   0 if successful, > 0 on error (value depends on error)
***/
static EMSTATUS display_temperature(int32_t tC) {
	uint32_t pwr10;                       /* which digit to print */
	EMSTATUS err;

	/* Maximum width is 7 characters, either 'XXX.X C' or '-XX.X C'.  As
	   with humidity, the first character will usually not print so put
	   it in the first column, apparently centering the value. */

	err = TEXTDISPLAY_WriteString(app.text, TEXTDISPLAY_ESC_SEQ_CURSOR_HOME_VT52);
	RET_ON_ERR(err);
	err = TEXTDISPLAY_WriteString(app.text, TEXTDISPLAY_ESC_SEQ_CURSOR_DOWN_ONE_LINE);
	RET_ON_ERR(err);

	if (tC < 0) {
		err = TEXTDISPLAY_WriteChar(app.text, '-');
		RET_ON_ERR(err);
		tC = -tC;
		/* Temp storage spec is -65 C, so don't worry about -100.  The minus
		   sign has replaced the 100's digit. */
		pwr10 = 10000;
	} else {
		pwr10 = 100000;
	}

	/* Round up the tenths digit. */
	tC += 50;

	while (10 < pwr10) {
		/* Don't print leading zeros before the unit's digit (pwr10==1000),
		   put do print it and after the decimal point. */
		if ((0 == (tC / pwr10)) && (1000 < pwr10)) {
			err = TEXTDISPLAY_WriteChar(app.text, ' ');
			RET_ON_ERR(err);
		} else {
			err = TEXTDISPLAY_WriteChar(app.text, '0' + (tC / pwr10));
			RET_ON_ERR(err);
		}
		if (1000 == pwr10) {
			err = TEXTDISPLAY_WriteChar(app.text, '.');
			RET_ON_ERR(err);
		}

		tC %= pwr10;
		pwr10 /= 10;
	}

	err = TEXTDISPLAY_WriteChar(app.text, ' ');
	RET_ON_ERR(err);

	err = TEXTDISPLAY_WriteChar(app.text, ';');
	RET_ON_ERR(err);

	return EMSTATUS_OK;
}

/***
    clear_graphs:  Erase any points already drawn in both sub-graphs.
***/
static void clear_graphs(void) {

	GRAPHDISPLAY_Clear(app.graph, 0, false);
	GRAPHDISPLAY_Clear(app.graph, 1, true);
}

/***
    graph_humidity:  Draw the last measurement in the log to the graph.
                     May re-scale and re-draw the entire log if necessary.
                     The log ID must point to the last sample.
    returns:   0 if successful, > 0 on error (value depends on error)
    modifies:  htlog
***/
static EMSTATUS graph_humidity(void) {
	uint8_t wgraph;                       /* sub-graph width */
	uint8_t x[NLOG], y[NLOG];             /* pixel coordinates in sub-graph */
	uint8_t npt;                          /* number of points to draw */
	bool clear;                           /* true to clear graph */
	uint8_t i;
	EMSTATUS err;

	/* Calculate the point(s) in one swoop so we only grab the lock once. */
	END_ON_LOCKERR(htlog_lock);

	if (0 == htlog.rh[htlog.id]) {
		UNLOCK(htlog_lock);
		return EMSTATUS_OK;
	}

	/* If graph bounds are correct then just draw the point.  Otherwise
	   must re-calculate the bounds, clear the graph, and draw all points. */
	if ((0 < htlog.g0min) && (htlog.g0min < htlog.rhmin) &&
	    (0 < htlog.g0scl) &&
	    ((htlog.rhmax - htlog.g0min) < (GRAPH_HT * htlog.g0scl))) {
		x[0] = htlog.id;
		y[0] = (((htlog.rh[htlog.id] - htlog.g0min) / 100) * GRAPH_HT) / htlog.g0scl;
		npt = 1;
		clear = false;
	} else {
		npt = 0;
		if ((htlog.rhmax - htlog.rhmin) < GRAPH_HT) {
			clear = false;
			/* Center range (no scaling) about average. */
			htlog.g0min = (((htlog.rhmin + htlog.rhmax - GRAPH_MINRANGE) / 2) / 100) * 100;
			htlog.g0scl = GRAPH_MINRANGE / GRAPH_HT;
		} else {
			clear = true;
			htlog.g0min = (htlog.rhmin / 100) * 100;
			htlog.g0scl = ((htlog.rhmax - htlog.g0min) + GRAPH_HT - 1) / GRAPH_HT;
			for (i=0; i<htlog.id; i++) {
				if (0 == htlog.rh[i]) {
					continue;
				}
				x[npt] = i;
				y[npt] = (((htlog.rh[i] - htlog.g0min) / 100) * GRAPH_HT) / htlog.g0scl;
				npt += 1;
			}
		}
		x[npt] = htlog.id;
		y[npt] = (((htlog.rh[htlog.id] - htlog.g0min) / 100) * GRAPH_HT) / htlog.g0scl;
		npt += 1;
	}

	END_ON_UNLOCKERR(htlog_lock);

	if (0 == npt) {
		return EMSTATUS_OK;
	}

	GRAPHDISPLAY_GetWidth(app.graph, &wgraph);
	if (wgraph <= x[npt-1]) {
		DPRINTF("too many points for graph (%d vs. w %d), need scrolling\n",
		        x[npt-1], wgraph);
		return EMSTATUS_FAIL;
	}

	if (clear) {
		err = GRAPHDISPLAY_Clear(app.graph, 0, false);
		RET_ON_ERR(err);
	}

	for (i=0; i<npt-1; i++) {
		err = draw_2x2point(0, x[i], y[i], false);
		RET_ON_ERR(err);
	}

	err = draw_2x2point(0, x[i], y[i], true);
	RET_ON_ERR(err);

	return EMSTATUS_OK;
onerror:
	return EMSTATUS_FAIL;
}

/***
    graph_temperature:  Draw the last measurement in the log to the graph.
                        May re-scale and re-draw the entire log if necessary.
                        The log ID must point to the last sample.
    returns:   0 if successful, > 0 on error (value depends on error)
    modifies:  htlog
***/
static EMSTATUS graph_temperature(void) {
	uint8_t wgraph;                       /* sub-graph width */
	uint8_t x[NLOG], y[NLOG];             /* pixel coordinates in sub-graph */
	uint8_t npt;                          /* number of points to draw */
	bool clear;                           /* true to clear graph */
	uint8_t i;
	EMSTATUS err;

	/* Calculate the point(s) in one swoop so we only need the lock once. */
	END_ON_LOCKERR(htlog_lock);

	if (0 == htlog.tC[htlog.id]) {
		UNLOCK(htlog_lock);
		return EMSTATUS_OK;
	}

	/* If graph bounds are correct then just draw the point.  Otherwise
	   must re-calculate the bounds, clear the graph, and draw all points. */
	if ((0 < htlog.g1min) && (htlog.g1min < htlog.tCmin) &&
	    (0 < htlog.g1scl) &&
	    ((htlog.tCmax - htlog.g1min) < (GRAPH_HT * htlog.g1scl))) {
		x[0] = htlog.id;
		y[0] = (((htlog.tC[htlog.id] - htlog.g1min) / 100) * GRAPH_HT) / htlog.g1scl;
		npt = 1;
		clear = false;
	} else {
		npt = 0;
		if ((htlog.tCmax - htlog.tCmin) < GRAPH_HT) {
			clear = false;
			/* Center range (no scaling) about average. */
			htlog.g1min = (((htlog.tCmin + htlog.tCmax - GRAPH_MINRANGE) / 2) / 100) * 100;
			htlog.g1scl = GRAPH_MINRANGE / GRAPH_HT;
		} else {
			clear = true;
			htlog.g1min = (htlog.tCmin / 100) * 100;
			htlog.g1scl = ((htlog.tCmax - htlog.g1min) + GRAPH_HT - 1) / GRAPH_HT;
			for (i=0; i<htlog.id; i++) {
				if (0 == htlog.tC[i]) {
					continue;
				}
				x[npt] = i;
				y[npt] = (((htlog.tC[i] - htlog.g1min) / 100) * GRAPH_HT) / htlog.g1scl;
				npt += 1;
			}
		}
		x[npt] = htlog.id;
		y[npt] = (((htlog.tC[htlog.id] - htlog.g1min) / 100) * GRAPH_HT) / htlog.g1scl;
		npt += 1;
	}

	END_ON_UNLOCKERR(htlog_lock);

	if (0 == npt) {
		return EMSTATUS_OK;
	}

	GRAPHDISPLAY_GetWidth(app.graph, &wgraph);
	if (wgraph <= x[npt-1]) {
		DPRINTF("too many points for graph (%d vs. width %d), need scrolling\n",
		        x[npt-1], wgraph);
		return EMSTATUS_FAIL;
	}

	if (clear) {
		err = GRAPHDISPLAY_Clear(app.graph, 1, false);
		RET_ON_ERR(err);
	}

	for (i=0; i<npt-1; i++) {
		err = draw_2x2point(1, x[i], y[i], false);
		RET_ON_ERR(err);
	}

	err = draw_2x2point(1, x[i], y[i], true);
	RET_ON_ERR(err);

	return EMSTATUS_OK;
onerror:
	return EMSTATUS_FAIL;
}

/***
    draw_2x2point:  Draw a 2x2 square on the LCD with the bottom right
                    corner at x,y.
    args:           graphID - 0 for left graph, 1 for right
                    x, y - lower right corner of point on screen
                    redraw - true to update display, false draws in memory
    returns:   0 if successful, > 0 on error (value depends on error)
***/
static EMSTATUS draw_2x2point(uint8_t graphID, uint8_t x, uint8_t y, bool redraw) {
	EMSTATUS err;

	/* We don't check if x, y legit because the graph routine does that.
	   Ignore OOB (ie. BADPARAM) errors. */

	err = GRAPHDISPLAY_SetPoint(app.graph, graphID, x, y, true, redraw);
	if ((EMSTATUS_OK != err) && (GRAPHDISPLAY_EMSTATUS_BADPARAM != err)) {
		return err;
	}

	err = GRAPHDISPLAY_SetPoint(app.graph, graphID, x-1, y, true, redraw);
	if ((EMSTATUS_OK != err) && (GRAPHDISPLAY_EMSTATUS_BADPARAM != err)) {
		return err;
	}

	err = GRAPHDISPLAY_SetPoint(app.graph, graphID, x, y-1, true, redraw);
	if ((EMSTATUS_OK != err) && (GRAPHDISPLAY_EMSTATUS_BADPARAM != err)) {
		return err;
	}

	err = GRAPHDISPLAY_SetPoint(app.graph, graphID, x-1, y-1, true, redraw);
	if ((EMSTATUS_OK != err) && (GRAPHDISPLAY_EMSTATUS_BADPARAM != err)) {
		return err;
	}

	return EMSTATUS_OK;
}


/**** Zwave Support For Sensor Readings (Sensor Multilevel CC) ****/

/***
    send_temperature:  If we're included in a network, send the temperature
                       to the root node.  Does nothing if we're not included
                       (not an error).
    args:              tC - temperature in milli-degrees C
    returns:   successful/failed/unsupported flag
***/
static received_frame_status_t send_temperature(int32_t tC) {

	if (!am_included()) {
		return EMSTATUS_OK;
	}

	return sendSensorValue(SENSOR_TEMP, tC, appSensorLevel(SENSOR_TEMP, SCALE_DEGC));
}

/***
    send_humidity:  If we're included in a network, send the humidity to the
                    root node.  Does nothing if we're not included (not an 
                    error).
    args:           rh - relative humidity, milli-percent
    returns:   successful/failed/unsupported flag
***/
static received_frame_status_t send_humidity(uint32_t rh) {

	if (!am_included()) {
		return EMSTATUS_OK;
	}

	return sendSensorValue(SENSOR_RH, rh, appSensorLevel(SENSOR_RH, SCALE_PCT));
}

/***
    appSensorValidType:  Validate the sensor type that's been passed with a
                         GET command.  Returns the argument if valid, or a
                         default (humidity sensor) if not.  Use the return
                         value for any other application calls.
    args:                type - SENSOR_* sensor type
    returns:   type or valid default value
***/
uint8_t appSensorValidType(uint8_t type) {

	switch (type) {
	case SENSOR_TEMP:     /* fall through */
	case SENSOR_RH:
		return type;
	default:
		return SENSOR_RH;
	}
}

/***
    appSensorMask:  Populate the masks array with the value bit masks we
                    support.  There's only one byte for the two - air
                    temperature (0x01) and humidity (0x10).
    args:           masks - supported sensor type bit masks
                    nmask - number of mask bytes we set
    modifies:  masks, nmask
***/
void appSensorMask(uint8_t *masks, uint8_t *nmask) {

	masks[0] = 0x11;
	*nmask = 1;
}

/***
    appSensorScaleMask:  Return the bit mask for the scales the given sensor
                         type supports.  If the sensor isn't supported then
                         return 0 (recommended behavior).
    args:                type - SENSOR_* sensor type
    returns:   bit mask
***/
uint8_t appSensorScaleMask(uint8_t type) {

	switch (type) {
	case SENSOR_TEMP:
		/* 0x01 for Celcius, 0x02 for Fahrenheit. */
		return 0x03;
	case SENSOR_RH:
		/* 0x01 for relative humidity. */
		return 0x01;
	default:
		/* Recommended behavior by spec. */
		return 0x00;
	}
}

/***
    appSensorValue:  Return the appropriate sensor reading.  The scale must
                     be unpacked from the commmand, ie. it's passed in bits
                     3 and 4 and the argument here must be in bits 0 and 1.
                     Values are always in milli-percent or milli-degrees.
                     The default for an invalid type is relative humidity,
                     for scale is temp C.  An error reading the sensor will
                     return a value of 0.
    args:            type - SENSOR_* sensor type
                     scale - SCALE_* of return value
    returns:   sensor reading
***/
int32_t appSensorValue(uint8_t type, uint8_t scale) {
	uint32_t rh;                          /* humidity measurement */
	int32_t tC;                           /* temperature measurement */
	uint32_t err;

	switch (type) {
	case SENSOR_TEMP:
		err = read_HT(NULL, &tC);
		if (0 != err) {
			return 0;
		}
		switch (scale) {
		case SCALE_DEGF:
			/* + 2 for round-off. */
			return (32000 + (((9 * tC) + 2) / 5));
		case SCALE_DEGC:     /* fall through */
		default:
			return tC;
		}
	case SENSOR_RH:        /* fall through */
	default:
		err = read_HT(&rh, NULL);
		if (0 != err) {
			return 0;
		}
		switch (scale) {
		case SCALE_PCT:       /* fall through */
		default:
			return rh;
		}
	}
}

/***
    appSensorLevel:  Return the precision/scale/size byte for _REPORT.
                     The scale must be unpacked from the GET command, ie.
                     it's passed in bits 3 and 4 and the argument here must
                     be shifted to 0 and 1.  The default type is relative
                     humidity, the default temp scale is C.
    args:           type - SENSOR_* type
                    scale - SCALE_* of return value
    returns:   precision/scale/size byte
***/
uint8_t appSensorLevel(uint8_t type, uint8_t scale) {

	switch (type) {
	case SENSOR_TEMP:
		switch (scale) {
		case SCALE_DEGF:
			return 0x6C;
		case SCALE_DEGC:     /* fall through */
		default:
			return 0x64;
		}
	case SENSOR_RH:        /* fall through */
	default:
		switch (scale) {
		case SCALE_PCT:      /* fall through */
		default:
			return 0x64;
		}
	}
}


/**** Zwave Support For Logging (Configuration CC) ****/

/***
    appConfigValidParam:  Return the argument if valid, a default parameter
                          (spec says lowest available ID, ie. 1) if not.
    args:                 paramID - parameter identifier
    returns:   valid identifer or default
***/
uint8_t appConfigValidParam(uint8_t paramID) {

	switch (paramID) {
	case CONFIG_TSAMPLE:     /* fall through */
	case CONFIG_NSAMPLE:     /* fall through */
	case CONFIG_RUN:
		return paramID;
	default:
		return CONFIG_TSAMPLE;
	}
}

/***
    appConfigValue:  Return the value of the logging parameter (for the run
                     state toggle this will be true if running, 0 if stopped).
                     The default for an invalid ID is the sample time.  The
                     actual size of the parameter may be less than 4 bytes.
    args:            paramID - parameter identifier
    returns:   value of parameter (or default if invalid)
***/
int32_t appConfigValue(uint8_t paramID) {
	int32_t val;                          /* value to return */

	END_ON_LOCKERR(htlog_lock);
	switch (paramID) {
	case CONFIG_NSAMPLE:
		val = htlog.nsample;
		break;
	case CONFIG_RUN:
		val = htlog.running;
		break;
	case CONFIG_TSAMPLE:
		val = htlog.tsample;
		break;
	default:
		val = 0;
		break;
	}
	END_ON_UNLOCKERR(htlog_lock);

	return val;

onerror:
	return 0;
}

/***
    appConfigLevel:  Return the packed size byte for the REPORT.  If paramID
                     is invalid returns the size of the default parameter.
    args:            paramID - parameter identifier
    returns:   'level' value specification
***/
uint8_t appConfigLevel(uint8_t paramID) {

	switch (paramID) {
	case CONFIG_NSAMPLE:     /* fall through */
	case CONFIG_RUN:
		return 1;
	case CONFIG_TSAMPLE:     /* fall through */
	default:
		return 4;
	}
}

/***
    appConfigSet:  Change the value of the parameter.  Clips values as
                   needed.  Ignores sampling parameter changes if the logger
                   is running.  The RUN parameter toggles the logger.
    args:          paramID - parameter identifier
                   val - new value
                   dflt - true to set parameter to default state
    returns:   true if parameter set, 0 if invalid parameter
    modifies:  htlog
***/
bool appConfigSet(uint8_t paramID, int32_t val, uint8_t dflt) {
	uint32_t err;

	END_ON_LOCKERR(htlog_lock);
	
	switch (paramID) {
	case CONFIG_NSAMPLE:
		if (htlog.running) {
			break;
		}
		if (dflt || (val < 0)) {
			htlog.nsample = 0;
		} else {
			htlog.nsample = val;
		}
		err = bound_nsample(&(htlog.nsample));
		if (EMSTATUS_OK != err) {
			UNLOCK(htlog_lock);
			return false;
		}
		DPRINTF(" set logger parameter nsample to %d\n", htlog.nsample);
		break;
	case CONFIG_TSAMPLE:
		if (htlog.running) {
			break;
		}
		if (dflt) {
			htlog.tsample = TSAMPLE;
		} else if (val <= 0) {
			htlog.tsample = 1;
		} else {
			htlog.tsample = val;
		}
		DPRINTF(" set logger parameter tsample to %d\n", htlog.tsample);
		break;
	case CONFIG_RUN:
		if (htlog.running) {
			if (!val) {
				DPRINTF(" stop logging\n");
				htlog.nsample = 0;
			} else {
				DPRINTF("ignoring log start because already running\n");
			}
		} else {
			if (0 == htlog.nsample) {
				DPRINTF(" must set number of samples before starting log\n");
			} else if (val) {
				DPRINTF(" start logging %d samples every %d sec\n",
				        htlog.nsample, htlog.tsample);
				END_ON_UNLOCKERR(htlog_lock);
				start_sampling();
				return true;
			} else {
				DPRINTF("ignoring log stop because already stopped\n");
			}
		}
		break;
	default:
		UNLOCK(htlog_lock);
		return false;
	}

	END_ON_UNLOCKERR(htlog_lock);
	return true;

onerror:
	return false;
}



/**** Other Hardware ****/

/***
    set_indicator:  Set our learn mode indicator LED (1) on (node included), 
                    off (excluded), or blinking (learn in progress).
    args:           state - learn mode outcome
***/
static void set_indicator(enum netstate state) {

	switch (state) {
	case LEARN_INCLUDED:
		/**!!!**/
		/* Can't figure out how to modify this to handle empty off periods, ie.
		   to leave LED on without using the timer.  The control will shut off
		   the LED, and we can manually turn it on/off.  It seems like the
		   timer is a temporary override - if the LED starts on (included), then
		   after the blinking for the learn it would stay on.
		*/
		Board_IndicatorControl(0, 0, 0, false);
		Board_SetLed(BOARD_LED1, LED_ON);
		break;
	case LEARN_EXCLUDED:
		Board_IndicatorControl(0, 0, 0, false);
		Board_SetLed(BOARD_LED1, LED_OFF);
		break;
	case LEARN_UNDERWAY:
		Board_IndicatorControl(100, 900, 0, false);
		break;
	case LEARN_FAILED:
		Board_IndicatorControl(100, 100, 75, false);
		break;
	default:
		DPRINTF("unknown network state %d\n", state);
		Board_IndicatorControl(0, 0, 0, false);
	}
}




/**** ZWave/ZAF Application ****/

/***
    ApplicationInit:  Initialize the hardware and register the main task with
                      ZAF.  Crashes through an assert if the main task does
                      not register.
    args:             reset - reason for initializing application
    returns:   APPLICATION_RUNNING
               < 0 on failure (value depends on error)
    modifies:  zwcfg, nif, radio, s2keys, nsample
***/
ZW_APPLICATION_STATUS ApplicationInit(EResetReason_t reset) {
	uint32_t err;

	AppTimerInit(EVT_TIMER, NULL);

	memset(dbgbuff, 0, sizeof_array(dbgbuff));
	memset(&zwcfg, 0, sizeof(zwcfg));

	htlog_lock = xSemaphoreCreateMutexStatic(&htlog_mutex);
	ASSERT(NULL != htlog_lock);

	/* Per ZW_application_transport_interface.h the pointers in the structure
	   must not change although their contents can. */

	copy_nif(&appnif1, &appnif2);

	radio.iListenBeforeTalkThreshold = ELISTENBEFORETALKTRESHOLD_DEFAULT;
	/* These values come from config_rf.h. */
	radio.iTxPowerLevelMax = MAX_TXPWR;
	radio.iTxPowerLevelAdjust = MEASURED_0DBM_TXPWR;
	radio.eRegion = RADIO_REGION;

	s2keys = 0;

	zwcfg.pVirtualSlaveNodeInfoTable = NULL;
	zwcfg.pSecureKeysRequested = &s2keys;
	zwcfg.pNodeInfo = &appnif2;
	zwcfg.pRadioConfig = &radio;

	clear_log();

	err = Board_Init();
	ASSERT(0 == err);

#ifdef DEBUGPRINT
	ZAF_UART0_enable(115200, true, false);
	DebugPrintConfig(dbgbuff, sizeof_array(dbgbuff), ZAF_UART0_tx_send);
	DPRINT("\n");
#endif   /* DEBUGPRINT*/

	err = BRD420xBoardInit(radio.eRegion);
	ASSERT(0 != err);
	err = setup_display();
	ASSERT(0 == err);

	CC_Version_SetApplicationVersionInfo(APP_VERSION, APP_REVISION, APP_PATCH, 0);

	err = ZW_ApplicationRegisterTask(ApplicationTask, EVT_ZWRX, EVT_ZWCMD, &zwcfg);
	ASSERT(err);

	return APPLICATION_RUNNING;
}

/***
    ApplicationTask:  The main application task.  Loops endlessly, handling
                      events from the framework, after configuring the 
                      hardware and setting up the event queue(s).
    args:             zwapp - ZWave configuration and queues
    modifies:  evtqueue
***/
static void ApplicationTask(SApplicationHandles *zwapp) {
	SEventDistributor evtdistrib;         /* event router */
	SQueueNotifying evtnotify;            /* event queue with notifications */
	StaticQueue_t evtstatic;              /* stack-allocated queue structure */
	enum zwevtIDs evtqueue_mem[MAXEVENT]; /* memory backing queue */
	EventDistributorEventHandler handlers[] = {
		AppTimerNotificationHandler,
		handle_frame,
		handle_cmdstatus,
		handle_event
	};		                                /* event handlers (one per evtID) */
	TaskHandle_t apptask;                 /* FreeRTOS handle for this task */
	uint32_t err;

	apptask = xTaskGetCurrentTaskHandle();
	AppTimerSetReceiverTask(apptask);
	ZAF_Init(apptask, zwapp, &zwcfg, NULL);

	err = Transport_OnApplicationInitSW(&appnif1, NULL);
	ASSERT(0 != err);

	Board_EnableButton(BOARD_BUTTON_PB1);
	Board_EnableButton(BOARD_BUTTON_PB2);
	Board_EnableButton(BOARD_BUTTON_PB3);
	Board_EnableButton(BOARD_BUTTON_PB4);

	Board_SetLed(BOARD_LED1, LED_OFF);
	Board_SetLed(BOARD_LED2, LED_OFF);
	Board_SetLed(BOARD_LED3, LED_OFF);
	Board_SetLed(BOARD_LED4, LED_OFF);

	Board_IndicatorInit(BOARD_LED1);

	setup_i2c();
	AppTimerRegister(&sampler, true, sample_ht);

	/* We never exit from this function so can use local variables rather than
	   globals, because they won't be going away. */
	evtqueue = xQueueCreateStatic(MAXEVENT, sizeof(evtqueue_mem[0]),
	                              (uint8_t *) evtqueue_mem, &evtstatic);
	QueueNotifyingInit(&evtnotify, evtqueue, apptask, EVT_APP);
	ZAF_EventHelperInit(&evtnotify);
	EventDistributorConfig(&evtdistrib, sizeof_array(handlers), handlers, NULL);

	/* The FreeRTOS task loop.  Do not exit without cleaning up. */
	DPRINT("AppTask - enter main loop\n");
	for (;;) {
		EventDistributorDistribute(&evtdistrib, 10, 0);
	}
}


/**** Event Handling - ZWave ****/

/***
    Transport_ApplicationCommandHandlerEx:  Incoming command class handler.
    args:                                   rxopt - receive options
                                            frame - frame buffer
                                            len - number of bytes in buffer
    returns:   result of command handler or RECEIVED_FRAME_STATUS_NO_SUPPORT
***/
received_frame_status_t
Transport_ApplicationCommandHandlerEx(RECEIVE_OPTIONS_TYPE_EX *rxopt,
                                      ZW_APPLICATION_TX_BUFFER *frame,
                                      uint8_t len) {
	received_frame_status_t status;       /* handler result */

	switch (frame->ZW_Common.cmdClass) {
	case COMMAND_CLASS_VERSION:
		status = CC_Version_handler(rxopt, frame, len);
		break;
	case COMMAND_CLASS_SENSOR_MULTILEVEL:
		status = handleCCMultilevelSensor(rxopt, frame, len);
		break;
	case COMMAND_CLASS_CONFIGURATION:
		status = handleCCConfiguration(rxopt, frame, len);
		break;
	default:
		DPRINTF("  no handler command class 0x%02x\n", frame->ZW_Common.cmdClass);
		status = RECEIVED_FRAME_STATUS_NO_SUPPORT;
		break;
	}

	return status;
}

/***
    handle_frame:  Event handler for ZWave frames.
***/
static void handle_frame(void) {
	SApplicationHandles *zwapp;           /* ZWave configuration and queues */
	QueueHandle_t queue;                  /* frame queue */
	SZwaveReceivePackage rxpkg;           /* incoming frame and meta-data */
	uint8_t *frame;                       /* raw payload */
	int i;

	zwapp = ZAF_getAppHandle();
	queue = zwapp->ZwRxQueue;

	while (pdTRUE == xQueueReceive(queue, (void *) &rxpkg, 0)) {
		switch (rxpkg.eReceiveType) {
		case EZWAVERECEIVETYPE_SINGLE:
			/* Show incoming packets with this. */
			/*
			DPRINTF("received frame type %3d (SINGLE)  len %d\n",
			        rxpkg.eReceiveType, rxpkg.uReceiveParams.Rx.iLength);
			frame = (uint8_t *) &(rxpkg.uReceiveParams.Rx.Payload);
			DPRINT("  0x");
			for (i=0; i<rxpkg.uReceiveParams.Rx.iLength; i++) {
				DPRINTF(" %02x", frame[i]);
				if (0 == ((i + 1) % 36)) {
					DPRINT("\n");
				}
			}
			if (0 != (i % 36)) {
				DPRINT("\n");
			}
			*/
			ZAF_CP_CommandPublish(ZAF_getCPHandle(), &rxpkg);
			break;
		case EZWAVERECEIVETYPE_MULTI:
			DPRINTF("received frame type %3d (MULTI)\n", rxpkg.eReceiveType);
			break;
		case EZWAVERECEIVETYPE_NODE_UPDATE:
			DPRINTF("received frame type %3d (UPDATE)\n", rxpkg.eReceiveType);
			break;
		case EZWAVERECEIVETYPE_SECURITY_EVENT:
			DPRINTF("received frame type %3d (SECURITY)\n", rxpkg.eReceiveType);
			break;
		default:
			DPRINTF("received frame type %3d (???)\n", rxpkg.eReceiveType);
		}
	}
	
}

/***
    handle_cmdstatus:  Event handler for ZAF commands.
    modifies:  app
***/
static void handle_cmdstatus(void) {
	SApplicationHandles *zwapp;           /* ZWave configuration and queues */
	QueueHandle_t queue;                  /* command status queue */
	SZwaveCommandStatusPackage cmdstatus; /* general command status */
	SZWaveTransmitStatus *txstatus;       /* specific status for queued item */
	void (*txcb)(uint8_t, TX_STATUS_TYPE *);  /* callback fn after transmit */
	enum netstate learned;                /* learn mode outcome */

	zwapp = ZAF_getAppHandle();
	queue = zwapp->ZwCommandStatusQueue;

	while (pdTRUE == xQueueReceive(queue, (void *) &cmdstatus, 0)) {
		switch (cmdstatus.eStatusType) {
		case EZWAVECOMMANDSTATUS_TX:
			/* Make sure the callback is run or the queue will fill up quickly. */
			txstatus = &cmdstatus.Content.TxStatus;
			if (txstatus->bIsTxFrameLegal && (NULL != txstatus->Handle)) {
				if (ZAF_TSE_TXCallback == txstatus->Handle) {
					ZAF_TSE_TXCallback(NULL);
				} else {
					txcb = txstatus->Handle;
					txcb(txstatus->TxStatus, &txstatus->ExtendedTxStatus);
				}
			}
			break;
		case EZWAVECOMMANDSTATUS_GENERATE_RANDOM:
			DPRINTF("!! implement command status type %3d (GEN RND)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_NODE_INFO:
			DPRINTF("!! implement command status type %3d (NIF)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_LEARN_MODE_STATUS:
			if (EINCLUSIONSTATE_EXCLUDED == zwapp->pNetworkInfo->eInclusionState) {
				learned = LEARN_EXCLUDED;
			} else {
				learned = LEARN_INCLUDED;
			}
			switch (cmdstatus.Content.LearnModeStatus.Status) {
			case ELEARNSTATUS_ASSIGN_COMPLETE:
				DPRINTF("command status type %3d (LEARN STATUS) - complete\n",
				        cmdstatus.eStatusType);
				app.learning = false;
				break;
			case ELEARNSTATUS_ASSIGN_NODEID_DONE:
				DPRINTF("command status type %3d (LEARN STATUS) - nodeID done\n",
				        cmdstatus.eStatusType);
				learned = LEARN_UNDERWAY;
				break;
			case ELEARNSTATUS_ASSIGN_RANGE_INFO_UPDATE:
				DPRINTF("command status type %3d (LEARN STATUS) - update range\n",
				        cmdstatus.eStatusType);
				learned = LEARN_UNDERWAY;
				break;
			case ELEARNSTATUS_ASSIGN_INFO_PENDING:
				DPRINTF("command status type %3d (LEARN STATUS) - pending info\n",
				        cmdstatus.eStatusType);
				learned = LEARN_UNDERWAY;
				break;
			case ELEARNSTATUS_ASSIGN_WAITING_FOR_FIND:
				DPRINTF("command status type %3d (LEARN STATUS) - wait for find\n",
				        cmdstatus.eStatusType);
				learned = LEARN_UNDERWAY;
				break;
			case ELEARNSTATUS_SMART_START_IN_PROGRESS:
				DPRINTF("command status type %3d (LEARN STATUS) - in smart start\n",
				        cmdstatus.eStatusType);
				learned = LEARN_UNDERWAY;
				break;
			case ELEARNSTATUS_LEARN_IN_PROGRESS:
				DPRINTF("command status type %3d (LEARN STATUS) - in progress\n",
				        cmdstatus.eStatusType);
				learned = LEARN_UNDERWAY;
				break;
			case ELEARNSTATUS_LEARN_MODE_COMPLETED_TIMEOUT:
				DPRINTF("command status type %3d (LEARN STATUS) - timed out\n",
				        cmdstatus.eStatusType);
				learned = LEARN_FAILED;
				app.learning = false;
				break;
			case ELEARNSTATUS_LEARN_MODE_COMPLETED_FAILED:
				DPRINTF("command status type %3d (LEARN STATUS) - failed\n",
				        cmdstatus.eStatusType);
				learned = LEARN_FAILED;
				app.learning = false;
				break;
			default:
				DPRINTF("command status type %3d (LEARN STATUS) - unknown status\n",
				        cmdstatus.eStatusType);
			}
			set_indicator(learned);
			break;
		case EZWAVECOMMANDSTATUS_SET_DEFAULT:
			DPRINTF("!! implement command status type %3d (SET DFLT)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_INVALID_TX_REQUEST:
			DPRINTF("!! implement command status type %3d (INVALID TX)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_INVALID_COMMAND:
			DPRINTF("!! implement command status type %3d (INVALID CMD)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_SET_RF_RECEIVE_MODE:
			DPRINTF("!! implement command status type %3d (RF RX MODE)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_IS_NODE_WITHIN_DIRECT_RANGE:
			DPRINTF("!! implement command status type %3d (DIRECT RNG)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_GET_NEIGHBOR_COUNT:
			DPRINTF("!! implement command status type %3d (NNEIGH)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_ARE_NODES_NEIGHBOURS:
			DPRINTF("!! implement command status type %3d (IS NEIGH)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_IS_FAILED_NODE_ID:
			DPRINTF("!! implement command status type %3d (IS FAILED)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_GET_ROUTING_TABLE_LINE:
			DPRINTF("!! implement command status type %3d (RT TBL)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_SET_ROUTING_INFO:
			DPRINTF("!! implement command status type %3d (SET RT)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_STORE_NODE_INFO:
			DPRINTF("!! implement command status type %3d (STORE NIF)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_GET_PRIORITY_ROUTE:
			DPRINTF("!! implement command status type %3d (GET PRIO RT)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_SET_PRIORITY_ROUTE:
			DPRINTF("!! implement command status type %3d (SET PRIO RT)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_SET_SLAVE_LEARN_MODE:
			DPRINTF("!! implement command status type %3d (SLAVE LRN)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_SET_SLAVE_LEARN_MODE_RESULT:
			DPRINTF("!! implement command status type %3d (SLAVE LRN RES)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_IS_VIRTUAL_NODE:
			DPRINTF("!! implement command status type %3d (IS VIRT)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_GET_VIRTUAL_NODES:
			DPRINTF("!! implement command status type %3d (GET VIRT)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_GET_CONTROLLER_CAPABILITIES:
			DPRINTF("!! implement command status type %3d (CTLR CAPAB)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_IS_PRIMARY_CTRL:
			DPRINTF("!! implement command status type %3d (IS PRIMARY)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_NETWORK_MANAGEMENT:
			DPRINTF("!! implement command status type %3d (NM)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_GET_BACKGROUND_RSSI:
			DPRINTF("!! implement command status type %3d (BG RSSI)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_AES_ECB:
			DPRINTF("!! implement command status type %3d (AES ECB)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_REMOVE_FAILED_NODE_ID:
			DPRINTF("!! implement command status type %3d (RM FAILED)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_REPLACE_FAILED_NODE_ID:
			DPRINTF("!! implement command status type %3d (REP FAILED)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_NETWORK_LEARN_MODE_START:
			app.learning = true;
			set_indicator(LEARN_UNDERWAY);
			break;
		case EZWAVECOMMANDSTATUS_ZW_SET_MAX_INCL_REQ_INTERVALS:
			DPRINTF("!! implement command status type %3d (MAX INTERVAL)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_NVM_BACKUP_RESTORE:
			DPRINTF("!! implement command status type %3d (NVM RESTORE)\n",
			        cmdstatus.eStatusType);
			break;
		case EZWAVECOMMANDSTATUS_PM_SET_POWERDOWN_CALLBACK:
			DPRINTF("!! implement command status type %3d (PWR CB)\n",
			        cmdstatus.eStatusType);
			break;
		default:
			DPRINTF("command status type %3d (???)\n", cmdstatus.eStatusType);
		}
	}

}


/**** Miscellaneous Functions ****/

/***
    am_included:  Check if we're in a network - if ZAF knows we have a
                  nodeID assigned.
    returns:   true if included, 0 if not
***/
static bool am_included(void) {
	SApplicationHandles *zwapp;           /* ZWave configuration and queues */

	zwapp = ZAF_getAppHandle();
	return (0 != zwapp->pNetworkInfo->NodeId);
}

/***
    copy_nif:  Copy the node information from one structure to another.
    args:      src - NIF to copy
               dest - second version of NIF
    modifies:  dest
***/
static void copy_nif(app_node_information_t *src, SAppNodeInfo_t *dest) {

	dest->CommandClasses.UnSecureIncludedCC.iListLength =
		src->cmdClassListNonSecureCount;
	dest->CommandClasses.UnSecureIncludedCC.pCommandClasses =
		src->cmdClassListNonSecure;
	dest->CommandClasses.SecureIncludedUnSecureCC.iListLength =
		src->cmdClassListNonSecureIncludedSecureCount;
	dest->CommandClasses.SecureIncludedUnSecureCC.pCommandClasses =
		src->cmdClassListNonSecureIncludedSecure;
	dest->CommandClasses.SecureIncludedSecureCC.iListLength =
		src->cmdClassListSecureCount;
	dest->CommandClasses.SecureIncludedSecureCC.pCommandClasses =
		src->cmdClassListSecure;
	dest->DeviceOptionsMask = src->deviceOptionsMask;
	dest->NodeType.generic = src->nodeType.generic;
	dest->NodeType.specific = src->nodeType.specific;
}
